import os
import ssl
import shutil
import socket
import logging
import collections

from OpenSSL import crypto  # type: ignore

import synapse.exc as s_exc
import synapse.common as s_common

defdir_default = '~/.syn/certs'
defdir = os.getenv('SYN_CERT_DIR')
if defdir is None:
    defdir = defdir_default

def iterFqdnUp(fqdn):
    levs = fqdn.split('.')
    for i in range(len(levs)):
        yield '.'.join(levs[i:])

TEN_YEARS = 10 * 365 * 24 * 60 * 60

logger = logging.getLogger(__name__)

class CertDir:
    '''
    Certificate loading/generation/signing utilities.

    Features:
        * Locates and load certificates, keys, and certificate signing requests (CSRs).
        * Generates keypairs for users, hosts, and certificate authorities (CAs), supports both signed and self-signed.
        * Generates certificate signing requests (CSRs) for users, hosts, and certificate authorities (CAs).
        * Signs certificate signing requests (CSRs).
        * Generates PKCS#12 archives for use in browser.

    Args:
        path (str): Optional path which can override the default path directory.

    Notes:
        * All certificates will be loaded from and written to ~/.syn/certs by default. Set the environment variable
          SYN_CERT_DIR to override.
        * All certificate generation methods create 4096 bit RSA keypairs.
        * All certificate signing methods use sha256 as the signature algorithm.
        * CertDir does not currently support signing CA CSRs.
    '''

    def __init__(self, path=None):
        self.crypto_numbits = 4096
        self.signing_digest = 'sha256'

        self.certdirs = []
        self.pathrefs = collections.defaultdict(int)

        if path is None:
            path = (defdir,)

        if not isinstance(path, (list, tuple)):
            path = (path,)

        for p in path:
            self.addCertPath(p)

    def addCertPath(self, *path):

        fullpath = s_common.genpath(*path)
        self.pathrefs[fullpath] += 1

        if self.pathrefs[fullpath] == 1:
            self.certdirs.append(fullpath)

    def delCertPath(self, *path):
        fullpath = s_common.gendir(*path)
        self.pathrefs[fullpath] -= 1
        if self.pathrefs[fullpath] <= 0:
            self.certdirs.remove(fullpath)
            self.pathrefs.pop(fullpath, None)

    def genCaCert(self, name, signas=None, outp=None, save=True):
        '''
        Generates a CA keypair.

        Args:
            name (str): The name of the CA keypair.
            signas (str): The CA keypair to sign the new CA with.
            outp (synapse.lib.output.Output): The output buffer.

        Examples:
            Make a CA named "myca":

                mycakey, mycacert = cdir.genCaCert('myca')

        Returns:
            ((OpenSSL.crypto.PKey, OpenSSL.crypto.X509)): Tuple containing the private key and certificate objects.
        '''
        pkey, cert = self._genBasePkeyCert(name)
        ext0 = crypto.X509Extension(b'basicConstraints', False, b'CA:TRUE')
        cert.add_extensions([ext0])

        if signas is not None:
            self.signCertAs(cert, signas)
        else:
            self.selfSignCert(cert, pkey)

        if save:

            keypath = self._savePkeyTo(pkey, 'cas', '%s.key' % name)
            if outp is not None:
                outp.printf('key saved: %s' % (keypath,))

            crtpath = self._saveCertTo(cert, 'cas', '%s.crt' % name)
            if outp is not None:
                outp.printf('cert saved: %s' % (crtpath,))

        return pkey, cert

    def genHostCert(self, name, signas=None, outp=None, csr=None, sans=None, save=True):
        '''
        Generates a host keypair.

        Args:
            name (str): The name of the host keypair.
            signas (str): The CA keypair to sign the new host keypair with.
            outp (synapse.lib.output.Output): The output buffer.
            csr (OpenSSL.crypto.PKey): The CSR public key when generating the keypair from a CSR.
            sans (list): List of subject alternative names.

        Examples:
            Make a host keypair named "myhost":

                myhostkey, myhostcert = cdir.genHostCert('myhost')

        Returns:
            ((OpenSSL.crypto.PKey, OpenSSL.crypto.X509)): Tuple containing the private key and certificate objects.
        '''
        pkey, cert = self._genBasePkeyCert(name, pkey=csr)

        ext_sans = {'DNS:' + name}
        if isinstance(sans, str):
            ext_sans = ext_sans.union(sans.split(','))
        ext_sans = ','.join(sorted(ext_sans))

        cert.add_extensions([
            crypto.X509Extension(b'nsCertType', False, b'server'),
            crypto.X509Extension(b'keyUsage', False, b'digitalSignature,keyEncipherment'),
            crypto.X509Extension(b'extendedKeyUsage', False, b'serverAuth'),
            crypto.X509Extension(b'basicConstraints', False, b'CA:FALSE'),
            crypto.X509Extension(b'subjectAltName', False, ext_sans.encode('utf-8')),
        ])

        if signas is not None:
            self.signCertAs(cert, signas)
        else:
            self.selfSignCert(cert, pkey)

        if save:
            if not pkey._only_public:
                keypath = self._savePkeyTo(pkey, 'hosts', '%s.key' % name)
                if outp is not None:
                    outp.printf('key saved: %s' % (keypath,))

            crtpath = self._saveCertTo(cert, 'hosts', '%s.crt' % name)
            if outp is not None:
                outp.printf('cert saved: %s' % (crtpath,))

        return pkey, cert

    def genHostCsr(self, name, outp=None):
        '''
        Generates a host certificate signing request.

        Args:
            name (str): The name of the host CSR.
            outp (synapse.lib.output.Output): The output buffer.

        Examples:
            Generate a CSR for the host key named "myhost":

                cdir.genHostCsr('myhost')

        Returns:
            bytes: The bytes of the CSR.
        '''
        return self._genPkeyCsr(name, 'hosts', outp=outp)

    def genUserCert(self, name, signas=None, outp=None, csr=None, save=True):
        '''
        Generates a user keypair.

        Args:
            name (str): The name of the user keypair.
            signas (str): The CA keypair to sign the new user keypair with.
            outp (synapse.lib.output.Output): The output buffer.
            csr (OpenSSL.crypto.PKey): The CSR public key when generating the keypair from a CSR.

        Examples:
            Generate a user cert for the user "myuser":

                myuserkey, myusercert = cdir.genUserCert('myuser')

        Returns:
            ((OpenSSL.crypto.PKey, OpenSSL.crypto.X509)): Tuple containing the key and certificate objects.
        '''
        pkey, cert = self._genBasePkeyCert(name, pkey=csr)

        cert.add_extensions([
            crypto.X509Extension(b'nsCertType', False, b'client'),
            crypto.X509Extension(b'keyUsage', False, b'digitalSignature'),
            crypto.X509Extension(b'extendedKeyUsage', False, b'clientAuth'),
            crypto.X509Extension(b'basicConstraints', False, b'CA:FALSE'),
        ])

        if signas is not None:
            self.signCertAs(cert, signas)
        else:
            self.selfSignCert(cert, pkey)

        if save:
            crtpath = self._saveCertTo(cert, 'users', '%s.crt' % name)
            if outp is not None:
                outp.printf('cert saved: %s' % (crtpath,))

            if not pkey._only_public:
                keypath = self._savePkeyTo(pkey, 'users', '%s.key' % name)
                if outp is not None:
                    outp.printf('key saved: %s' % (keypath,))

        return pkey, cert

    def genClientCert(self, name, outp=None):
        '''
        Generates a user PKCS #12 archive.
        Please note that the resulting file will contain private key material.

        Args:
            name (str): The name of the user keypair.
            outp (synapse.lib.output.Output): The output buffer.

        Examples:
            Make the PKC12 object for user "myuser":

                myuserpkcs12 = cdir.genClientCert('myuser')

        Returns:
            OpenSSL.crypto.PKCS12: The PKCS #12 archive.
        '''
        ucert = self.getUserCert(name)
        if not ucert:
            raise s_exc.NoSuchFile(mesg='missing User cert', name=name)

        capath = self._getCaPath(ucert)
        cacert = self._loadCertPath(capath)
        if not cacert:
            raise s_exc.NoSuchFile(mesg='missing CA cert', path=capath)

        ukey = self.getUserKey(name)
        if not ukey:
            raise s_exc.NoSuchFile(mesg='missing User private key', name=name)

        ccert = crypto.PKCS12()
        ccert.set_friendlyname(name.encode('utf-8'))
        ccert.set_ca_certificates([cacert])
        ccert.set_certificate(ucert)
        ccert.set_privatekey(ukey)

        crtpath = self._saveP12To(ccert, 'users', '%s.p12' % name)
        if outp is not None:
            outp.printf('client cert saved: %s' % (crtpath,))

    def valUserCert(self, byts, cacerts=None):
        '''
        Validate the PEM encoded x509 user certificate bytes and return it.

        Args:
            byts (bytes): The bytes for the User Certificate.
            cacerts (tuple): A tuple of OpenSSL.crypto.X509 CA Certificates.

        Raises:
            OpenSSL.crypto.X509StoreContextError: If the certificate is not valid.

        Returns:
            OpenSSL.crypto.X509: The certificate, if it is valid.

        '''
        cert = crypto.load_certificate(crypto.FILETYPE_PEM, byts)

        if cacerts is None:
            cacerts = self.getCaCerts()

        store = crypto.X509Store()
        [store.add_cert(cacert) for cacert in cacerts]

        ctx = crypto.X509StoreContext(store, cert)
        ctx.verify_certificate()  # raises X509StoreContextError if unable to verify
        return cert

    def genUserCsr(self, name, outp=None):
        '''
        Generates a user certificate signing request.

        Args:
            name (str): The name of the user CSR.
            outp (synapse.lib.output.Output): The output buffer.

        Examples:
            Generate a CSR for the user "myuser":

                cdir.genUserCsr('myuser')

        Returns:
            bytes: The bytes of the CSR.
        '''
        return self._genPkeyCsr(name, 'users', outp=outp)

    def getCaCert(self, name):
        '''
        Loads the X509 object for a given CA.

        Args:
            name (str): The name of the CA keypair.

        Examples:
            Get the certificate for the  CA "myca"

                mycacert = cdir.getCaCert('myca')

        Returns:
            OpenSSL.crypto.X509: The certificate, if exists.
        '''
        return self._loadCertPath(self.getCaCertPath(name))

    def getCaCerts(self):
        '''
        Return a list of CA certs from the CertDir.

        Returns:
            [OpenSSL.crypto.X509]: List of CA certificates.
        '''
        retn = []

        for cdir in self.certdirs:

            path = s_common.genpath(cdir, 'cas')

            for name in os.listdir(path):
                if not name.endswith('.crt'):
                    continue

                full = s_common.genpath(cdir, 'cas', name)
                retn.append(self._loadCertPath(full))

        return retn

    def getCaCertPath(self, name):
        '''
        Gets the path to a CA certificate.

        Args:
            name (str): The name of the CA keypair.

        Examples:
            Get the path to the CA certificate for the CA "myca":

                mypath = cdir.getCACertPath('myca')

        Returns:
            str: The path if exists.
        '''
        for cdir in self.certdirs:
            path = s_common.genpath(cdir, 'cas', '%s.crt' % name)
            if os.path.isfile(path):
                return path

    def getCaKey(self, name):
        '''
        Loads the PKey object for a given CA keypair.

        Args:
            name (str): The name of the CA keypair.

        Examples:
            Get the private key for the CA "myca":

                mycakey = cdir.getCaKey('myca')

        Returns:
            OpenSSL.crypto.PKey: The private key, if exists.
        '''
        return self._loadKeyPath(self.getCaKeyPath(name))

    def getCaKeyPath(self, name):
        '''
        Gets the path to a CA key.

        Args:
            name (str): The name of the CA keypair.

        Examples:
            Get the path to the private key for the CA "myca":

                mypath = cdir.getCAKeyPath('myca')

        Returns:
            str: The path if exists.
        '''
        for cdir in self.certdirs:
            path = s_common.genpath(cdir, 'cas', '%s.key' % name)
            if os.path.isfile(path):
                return path

    def getClientCert(self, name):
        '''
        Loads the PKCS12 archive object for a given user keypair.

        Args:
            name (str): The name of the user keypair.

        Examples:
            Get the PKCS12 object for the user "myuser":

                mypkcs12 = cdir.getClientCert('myuser')

        Notes:
            The PKCS12 archive will contain private key material if it was created with CertDir or the easycert tool

        Returns:
            OpenSSL.crypto.PKCS12: The PKCS12 archive, if exists.
        '''
        return self._loadP12Path(self.getClientCertPath(name))

    def getClientCertPath(self, name):
        '''
        Gets the path to a client certificate.

        Args:
            name (str): The name of the client keypair.

        Examples:
            Get the path to the client certificate for "myuser":

                mypath = cdir.getClientCertPath('myuser')

        Returns:
            str: The path if exists.
        '''
        for cdir in self.certdirs:
            path = s_common.genpath(cdir, 'users', '%s.p12' % name)
            if os.path.isfile(path):
                return path

    def getHostCaPath(self, name):
        '''
        Gets the path to the CA certificate that issued a given host keypair.

        Args:
            name (str): The name of the host keypair.

        Examples:
            Get the path to the CA cert which issue the cert for "myhost":

                mypath = cdir.getHostCaPath('myhost')

        Returns:
            str: The path if exists.
        '''
        cert = self.getHostCert(name)
        if cert is None:
            return None

        return self._getCaPath(cert)

    def getHostCert(self, name):
        '''
        Loads the X509 object for a given host keypair.

        Args:
            name (str): The name of the host keypair.

        Examples:
            Get the certificate object for the host "myhost":

                myhostcert = cdir.getHostCert('myhost')

        Returns:
            OpenSSL.crypto.X509: The certificate, if exists.
        '''
        return self._loadCertPath(self.getHostCertPath(name))

    def getHostCertPath(self, name):
        '''
        Gets the path to a host certificate.

        Args:
            name (str): The name of the host keypair.

        Examples:
            Get the path to the host certificate for the host "myhost":

                mypath = cdir.getHostCertPath('myhost')

        Returns:
            str: The path if exists.
        '''
        for cdir in self.certdirs:
            path = s_common.genpath(cdir, 'hosts', '%s.crt' % name)
            if os.path.isfile(path):
                return path

    def getHostKey(self, name):
        '''
        Loads the PKey object for a given host keypair.

        Args:
            name (str): The name of the host keypair.

        Examples:
            Get the private key object for the host "myhost":

                myhostkey = cdir.getHostKey('myhost')

        Returns:
            OpenSSL.crypto.PKey: The private key, if exists.
        '''
        return self._loadKeyPath(self.getHostKeyPath(name))

    def getHostKeyPath(self, name):
        '''
        Gets the path to a host key.

        Args:
            name (str): The name of the host keypair.

        Examples:
            Get the path to the host key for the host "myhost":

                mypath = cdir.getHostKeyPath('myhost')

        Returns:
            str: The path if exists.
        '''
        for cdir in self.certdirs:
            path = s_common.genpath(cdir, 'hosts', '%s.key' % name)
            if os.path.isfile(path):
                return path

    def getUserCaPath(self, name):
        '''
        Gets the path to the CA certificate that issued a given user keypair.

        Args:
            name (str): The name of the user keypair.

        Examples:
            Get the path to the CA cert which issue the cert for "myuser":

                mypath = cdir.getUserCaPath('myuser')

        Returns:
            str: The path if exists.
        '''
        cert = self.getUserCert(name)
        if cert is None:
            return None

        return self._getCaPath(cert)

    def getUserCert(self, name):
        '''
        Loads the X509 object for a given user keypair.

        Args:
            name (str): The name of the user keypair.

        Examples:
            Get the certificate object for the user "myuser":

                myusercert = cdir.getUserCert('myuser')

        Returns:
            OpenSSL.crypto.X509: The certificate, if exists.
        '''
        return self._loadCertPath(self.getUserCertPath(name))

    def getUserCertPath(self, name):
        '''
        Gets the path to a user certificate.

        Args:
            name (str): The name of the user keypair.

        Examples:
            Get the path for the user cert for "myuser":

                mypath = cdir.getUserCertPath('myuser')

        Returns:
            str: The path if exists.
        '''
        for cdir in self.certdirs:
            path = s_common.genpath(cdir, 'users', '%s.crt' % name)
            if os.path.isfile(path):
                return path

    def getUserForHost(self, user, host):
        '''
        Gets the name of the first existing user cert for a given user and host.

        Args:
            user (str): The name of the user.
            host (str): The name of the host.

        Examples:
            Get the name for the "myuser" user cert at "cool.vertex.link":

                usercertname = cdir.getUserForHost('myuser', 'cool.vertex.link')

        Returns:
            str: The cert name, if exists.
        '''
        for name in iterFqdnUp(host):
            usercert = '%s@%s' % (user, name)
            if self.isUserCert(usercert):
                return usercert

    def getUserKey(self, name):
        '''
        Loads the PKey object for a given user keypair.


        Args:
            name (str): The name of the user keypair.

        Examples:
            Get the key object for the user key for "myuser":

                myuserkey = cdir.getUserKey('myuser')

        Returns:
            OpenSSL.crypto.PKey: The private key, if exists.
        '''
        return self._loadKeyPath(self.getUserKeyPath(name))

    def getUserKeyPath(self, name):
        '''
        Gets the path to a user key.

        Args:
            name (str): The name of the user keypair.

        Examples:
            Get the path to the user key for "myuser":

                mypath = cdir.getUserKeyPath('myuser')

        Returns:
            str: The path if exists.
        '''
        for cdir in self.certdirs:
            path = s_common.genpath(cdir, 'users', '%s.key' % name)
            if os.path.isfile(path):
                return path

    def importFile(self, path, mode, outp=None):
        '''
        Imports certs and keys into the Synapse cert directory

        Args:
            path (str): The path of the file to be imported.
            mode (str): The certdir subdirectory to import the file into.

        Examples:
            Import CA certifciate 'mycoolca.crt' to the 'cas' directory.

                certdir.importFile('mycoolca.crt', 'cas')

        Notes:
            importFile does not perform any validation on the files it imports.

        Returns:
            None
        '''
        if not os.path.isfile(path):
            raise s_exc.NoSuchFile(mesg=f'File {path} does not exist', path=path)

        fname = os.path.split(path)[1]
        parts = fname.rsplit('.', 1)
        ext = parts[1] if len(parts) == 2 else None

        if not ext or ext not in ('crt', 'key', 'p12'):
            mesg = 'importFile only supports .crt, .key, .p12 extensions'
            raise s_exc.BadFileExt(mesg=mesg, ext=ext)

        newpath = s_common.genpath(self.certdirs[0], mode, fname)
        if os.path.isfile(newpath):
            raise s_exc.FileExists(mesg=f'File {newpath} already exists', path=path)

        s_common.gendir(os.path.dirname(newpath))

        shutil.copy(path, newpath)
        if outp is not None:
            outp.printf('copied %s to %s' % (path, newpath))

    def isCaCert(self, name):
        '''
        Checks if a CA certificate exists.

        Args:
            name (str): The name of the CA keypair.

        Examples:
            Check if the CA certificate for "myca" exists:

                exists = cdir.isCaCert('myca')

        Returns:
            bool: True if the certificate is present, False otherwise.
        '''
        return self.getCaCertPath(name) is not None

    def isClientCert(self, name):
        '''
        Checks if a user client certificate (PKCS12) exists.

        Args:
            name (str): The name of the user keypair.

        Examples:
            Check if the client certificate "myuser" exists:

                exists = cdir.isClientCert('myuser')

        Returns:
            bool: True if the certificate is present, False otherwise.
        '''
        crtpath = self._getPathJoin('users', '%s.p12' % name)
        return os.path.isfile(crtpath)

    def isHostCert(self, name):
        '''
        Checks if a host certificate exists.

        Args:
            name (str): The name of the host keypair.

        Examples:
            Check if the host cert "myhost" exists:

                exists = cdir.isUserCert('myhost')

        Returns:
            bool: True if the certificate is present, False otherwise.
        '''
        return self.getHostCertPath(name) is not None

    def isUserCert(self, name):
        '''
        Checks if a user certificate exists.

        Args:
            name (str): The name of the user keypair.

        Examples:
            Check if the user cert "myuser" exists:

                exists = cdir.isUserCert('myuser')

        Returns:
            bool: True if the certificate is present, False otherwise.
        '''
        return self.getUserCertPath(name) is not None

    def signCertAs(self, cert, signas):
        '''
        Signs a certificate with a CA keypair.

        Args:
            cert (OpenSSL.crypto.X509): The certificate to sign.
            signas (str): The CA keypair name to sign the new keypair with.

        Examples:
            Sign a certificate with the CA "myca":

                cdir.signCertAs(mycert, 'myca')

        Returns:
            None
        '''
        cakey = self.getCaKey(signas)
        if cakey is None:
            raise s_exc.NoCertKey(mesg=f'Missing .key for {signas}')
        cacert = self.getCaCert(signas)
        if cacert is None:
            raise s_exc.NoCertKey(mesg=f'Missing .crt for {signas}')

        cert.set_issuer(cacert.get_subject())
        cert.sign(cakey, self.signing_digest)

    def signHostCsr(self, xcsr, signas, outp=None, sans=None):
        '''
        Signs a host CSR with a CA keypair.

        Args:
            cert (OpenSSL.crypto.X509Req): The certificate signing request.
            signas (str): The CA keypair name to sign the CSR with.
            outp (synapse.lib.output.Output): The output buffer.
            sans (list): List of subject alternative names.

        Examples:
            Sign a host key with the CA "myca":

                cdir.signHostCsr(mycsr, 'myca')

        Returns:
            ((OpenSSL.crypto.PKey, OpenSSL.crypto.X509)):  Tuple containing the public key and certificate objects.
        '''
        pkey = xcsr.get_pubkey()
        name = xcsr.get_subject().CN
        return self.genHostCert(name, csr=pkey, signas=signas, outp=outp, sans=sans)

    def selfSignCert(self, cert, pkey):
        '''
        Self-sign a certificate.

        Args:
            cert (OpenSSL.crypto.X509): The certificate to sign.
            pkey (OpenSSL.crypto.PKey): The PKey with which to sign the certificate.

        Examples:
            Sign a given certificate with a given private key:

                cdir.selfSignCert(mycert, myotherprivatekey)

        Returns:
            None
        '''
        cert.set_issuer(cert.get_subject())
        cert.sign(pkey, self.signing_digest)

    def signUserCsr(self, xcsr, signas, outp=None):
        '''
        Signs a user CSR with a CA keypair.

        Args:
            cert (OpenSSL.crypto.X509Req): The certificate signing request.
            signas (str): The CA keypair name to sign the CSR with.
            outp (synapse.lib.output.Output): The output buffer.

        Examples:
            cdir.signUserCsr(mycsr, 'myca')

        Returns:
            ((OpenSSL.crypto.PKey, OpenSSL.crypto.X509)): Tuple containing the public key and certificate objects.
        '''
        pkey = xcsr.get_pubkey()
        name = xcsr.get_subject().CN
        return self.genUserCert(name, csr=pkey, signas=signas, outp=outp)

    def _loadCasIntoSSLContext(self, ctx):

        for cdir in self.certdirs:

            path = s_common.genpath(cdir, 'cas')
            if not os.path.isdir(path):
                continue

            for name in os.listdir(path):
                if name.endswith('.crt'):
                    ctx.load_verify_locations(os.path.join(path, name))

    def getClientSSLContext(self, certname=None):
        '''
        Returns an ssl.SSLContext appropriate for initiating a TLS session

        Args:
            certname:   If specified, use the user certificate with the matching
                        name to authenticate to the remote service.
        '''
        sslctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
        self._loadCasIntoSSLContext(sslctx)

        if certname is not None:

            username = certname
            if username.find('@') != -1:
                user, host = username.split('@', 1)
                username = self.getUserForHost(user, host)

            if username is None:
                mesg = f'User certificate not found: {certname}'
                raise s_exc.NoSuchCert(mesg=mesg)

            certpath = self.getUserCertPath(username)
            if certpath is None:
                mesg = f'User certificate not found: {certname}'
                raise s_exc.NoSuchCert(mesg=mesg)

            keypath = self.getUserKeyPath(username)
            if keypath is None:
                mesg = f'User private key not found: {certname}'
                raise s_exc.NoCertKey(mesg=mesg)

            sslctx.load_cert_chain(certpath, keypath)

        return sslctx

    def getServerSSLContext(self, hostname=None, caname=None):
        '''
        Returns an ssl.SSLContext appropriate to listen on a socket

        Args:

            hostname:  if None, the value from socket.gethostname is used to find the key in the servers directory.
                       This name should match the not-suffixed part of two files ending in .key and .crt in the hosts
                       subdirectory

            caname: If not None, the given name is used to locate a CA certificate used to validate client SSL certs.

        '''
        if hostname is not None and hostname.find(',') != -1:
            # multi-hostname SNI routing has been requested
            ctxs = {}
            names = hostname.split(',')
            for name in names:
                ctxs[name] = self._getServerSSLContext(name, caname=caname)

            def snifunc(sslsock, sslname, origctx):
                sslsock.context = ctxs.get(sslname, origctx)
                return None

            sslctx = ctxs.get(names[0])
            sslctx.sni_callback = snifunc
            return sslctx

        return self._getServerSSLContext(hostname=hostname, caname=caname)

    def _getServerSSLContext(self, hostname=None, caname=None):
        sslctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        if hostname is None:
            hostname = socket.gethostname()

        certfile = self.getHostCertPath(hostname)
        if certfile is None:
            mesg = f'Missing TLS certificate file for host: {hostname}'
            raise s_exc.NoCertKey(mesg=mesg)

        keyfile = self.getHostKeyPath(hostname)
        if keyfile is None:
            mesg = f'Missing TLS key file for host: {hostname}'
            raise s_exc.NoCertKey(mesg=mesg)

        sslctx.load_cert_chain(certfile, keyfile)

        if caname is not None:
            cafile = self.getCaCertPath(caname)
            sslctx.verify_mode = ssl.VerifyMode.CERT_REQUIRED
            sslctx.load_verify_locations(cafile=cafile)

        return sslctx

    def saveCertPem(self, cert, path):
        '''
        Save a certificate in PEM format to a file outside the certdir.
        '''
        with s_common.genfile(path) as fd:
            fd.truncate(0)
            fd.write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert))

    def savePkeyPem(self, pkey, path):
        '''
        Save a private key in PEM format to a file outside the certdir.
        '''
        with s_common.genfile(path) as fd:
            fd.truncate(0)
            fd.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, pkey))

    def _checkDupFile(self, path):
        if os.path.isfile(path):
            raise s_exc.DupFileName(mesg=f'Duplicate file {path}', path=path)

    def _genBasePkeyCert(self, name, pkey=None):

        if pkey is None:
            pkey = crypto.PKey()
            pkey.generate_key(crypto.TYPE_RSA, self.crypto_numbits)

        cert = crypto.X509()
        cert.set_pubkey(pkey)
        cert.set_version(2)

        cert.gmtime_adj_notBefore(0)
        cert.gmtime_adj_notAfter(TEN_YEARS)  # Certpairs are good for 10 years

        cert.set_serial_number(int(s_common.guid(), 16))
        cert.get_subject().CN = name

        return pkey, cert

    def _genPkeyCsr(self, name, mode, outp=None):
        pkey = crypto.PKey()
        pkey.generate_key(crypto.TYPE_RSA, self.crypto_numbits)

        xcsr = crypto.X509Req()
        xcsr.get_subject().CN = name

        xcsr.set_pubkey(pkey)
        xcsr.sign(pkey, self.signing_digest)

        keypath = self._savePkeyTo(pkey, mode, '%s.key' % name)
        if outp is not None:
            outp.printf('key saved: %s' % (keypath,))

        csrpath = self._getPathJoin(mode, '%s.csr' % name)
        self._checkDupFile(csrpath)

        byts = crypto.dump_certificate_request(crypto.FILETYPE_PEM, xcsr)

        with s_common.genfile(csrpath) as fd:
            fd.truncate(0)
            fd.write(byts)

        if outp is not None:
            outp.printf('csr saved: %s' % (csrpath,))

        return byts

    def _getCaPath(self, cert):
        subj = cert.get_issuer()
        return self.getCaCertPath(subj.CN)

    def _getPathBytes(self, path):
        if path is None:
            return None
        return s_common.getbytes(path)

    def _getPathJoin(self, *paths):
        return s_common.genpath(self.certdirs[0], *paths)

    def _loadCertPath(self, path):
        byts = self._getPathBytes(path)
        if byts:
            return crypto.load_certificate(crypto.FILETYPE_PEM, byts)

    def _loadCsrPath(self, path):
        byts = self._getPathBytes(path)
        if byts:
            return self._loadCsrByts(byts)

    def _loadCsrByts(self, byts):
        return crypto.load_certificate_request(crypto.FILETYPE_PEM, byts)

    def _loadKeyPath(self, path):
        byts = self._getPathBytes(path)
        if byts:
            return crypto.load_privatekey(crypto.FILETYPE_PEM, byts)

    def _loadP12Path(self, path):
        byts = self._getPathBytes(path)
        if byts:
            # This API is deprecrated by PyOpenSSL and will need to be rewritten if pyopenssl is
            # updated from v21.x.x. The APIs that use this are not directly exposed via the
            # easycert tool currently, and are only used in unit tests.
            return crypto.load_pkcs12(byts)

    def _saveCertTo(self, cert, *paths):
        path = self._getPathJoin(*paths)
        self._checkDupFile(path)

        with s_common.genfile(path) as fd:
            fd.truncate(0)
            fd.write(self._certToByts(cert))

        return path

    def _certToByts(self, cert):
        return crypto.dump_certificate(crypto.FILETYPE_PEM, cert)

    def _pkeyToByts(self, pkey):
        return crypto.dump_privatekey(crypto.FILETYPE_PEM, pkey)

    def _savePkeyTo(self, pkey, *paths):
        path = self._getPathJoin(*paths)
        self._checkDupFile(path)

        with s_common.genfile(path) as fd:
            fd.truncate(0)
            fd.write(self._pkeyToByts(pkey))

        return path

    def _saveP12To(self, cert, *paths):
        path = self._getPathJoin(*paths)
        self._checkDupFile(path)

        with s_common.genfile(path) as fd:
            fd.truncate(0)
            fd.write(cert.export())

        return path

certdir = CertDir()
def getCertDir() -> CertDir:
    '''
    Get the singleton CertDir instance.

    Returns:
        CertDir: A certdir object.
    '''
    return certdir

def addCertPath(path):
    return certdir.addCertPath(path)

def delCertPath(path):
    return certdir.delCertPath(path)

def getCertDirn() -> str:
    '''
    Get the expanded default path used by the singleton CertDir instance.

    Returns:
        str: The path string.
    '''
    return s_common.genpath(defdir)



import synapse.exc as s_exc
import synapse.common as s_common
import synapse.tests.utils as s_test

class StormWhoisTest(s_test.SynTest):

    async def test_storm_whois_guid(self):

        async with self.getTestCore() as core:
            # IP netblock record
            props = {
                'net4': '10.0.0.0/28',
                'asof': 2554869000000,
                'id': 'NET-10-0-0-0-1',
                'status': 'validated',
            }
            stormcmd = '''
                [(inet:whois:iprec=$lib.inet.whois.guid($props, "iprec")
                    :net4=$props.net4
                    :asof=$props.asof
                    :id=$props.id
                    :status=$props.status)]
            '''
            opts = {'vars': {'props': props}}
            _ = await core.nodes(stormcmd, opts=opts)
            guid_exp = s_common.guid(sorted((props['net4'], str(props['asof']), props['id'])))
            self.len(1, await core.nodes(f'inet:whois:iprec={guid_exp}'))

            stormcmd = '''$props=$lib.dict(net4="10.0.0.0/28", asof=(2554869000000), id='NET-10-0-0-0-1', status=validated)
            return ($lib.inet.whois.guid($props, 'iprec'))
            '''
            guid = await core.callStorm(stormcmd)
            self.eq(guid_exp, guid)

            # contact
            pscontact = s_common.guid()
            props = {
                'contact': pscontact,
                'asof': 2554869000000,
                'roles': ('abuse', 'technical', 'administrative'),
                'asn': 123456,
                'id': 'SPM-3',
                'links': ('http://myrdap.com/SPM3',),
                'status': 'active',
            }
            stormcmd = '''
                [(inet:whois:ipcontact=$lib.inet.whois.guid($props, "ipcontact")
                    :contact=$props.contact
                    :asof=$props.asof
                    :id=$props.id
                    :status=$props.status)]
            '''
            opts = {'vars': {'props': props}}
            _ = await core.nodes(stormcmd, opts=opts)
            guid_exp = s_common.guid(sorted((props['contact'], str(props['asof']), props['id'])))
            self.len(1, await core.nodes(f'inet:whois:ipcontact={guid_exp}'))

            # query
            props = {
                'time': 2554869000000,
                'url': 'http://myrdap/rdap/?query=3300%3A100%3A1%3A%3Affff',
                'ipv6': '3300:100:1::ffff',
                'success': False,
            }
            stormcmd = '''
                [(inet:whois:ipquery=$lib.inet.whois.guid($props, "ipquery")
                    :time=$props.time
                    :url=$props.url
                    :ipv6=$props.ipv6
                    :success=$props.success)]
            '''
            opts = {'vars': {'props': props}}
            _ = await core.nodes(stormcmd, opts=opts)
            guid_exp = s_common.guid(sorted((str(props['time']), props['url'], props['ipv6'])))
            self.len(1, await core.nodes(f'inet:whois:ipquery={guid_exp}'))

            # Random guid cases
            props = {
                'fqdn': 'foo.bar',
            }
            stormcmd = '''
                [(inet:whois:ipquery=$lib.inet.whois.guid($props, "ipquery")
                    :fqdn=$props.fqdn)]
            '''
            opts = {'vars': {'props': props}}
            mesgs = await core.stormlist(stormcmd, opts=opts)
            warn = [m[1]['mesg'] for m in mesgs if m[0] == 'warn']
            self.isin('Insufficient guid vals identified, using random guid:', warn[0])
            self.len(1, await core.nodes(f'inet:whois:ipquery:fqdn={props["fqdn"]}'))

            props = {
                'asn': 9999,
            }
            stormcmd = '''
                [(inet:whois:ipcontact=$lib.inet.whois.guid($props, "ipcontact")
                    :asn=$props.asn)]
            '''
            opts = {'vars': {'props': props}}
            mesgs = await core.stormlist(stormcmd, opts=opts)
            warn = [m[1]['mesg'] for m in mesgs if m[0] == 'warn']
            self.isin('Insufficient guid vals identified, using random guid:', warn[0])
            self.len(1, await core.nodes(f'inet:whois:ipcontact:asn={props["asn"]}'))

            # Failure cases
            stormcmd = '''
                [(inet:whois:ipcontact=$lib.inet.whois.guid($props, "foobar"))]
            '''
            opts = {'vars': {'props': {}}}
            await self.asyncraises(s_exc.StormRuntimeError, core.nodes(stormcmd, opts=opts))

            stormcmd = '''
                [(inet:whois:ipcontact=$lib.inet.whois.guid($props, "ipcontact"))]
            '''
            opts = {'vars': {'props': 123}}
            await self.asyncraises(s_exc.StormRuntimeError, core.nodes(stormcmd, opts=opts))



import synapse.tests.utils as s_t_utils

import synapse.lookup.iso3166 as s_l_country


class CountryLookTest(s_t_utils.SynTest):

    def test_lookup_countries(self):
        self.eq(s_l_country.country2iso.get('united states of america'), 'us')
        self.eq(s_l_country.country2iso.get('mexico'), 'mx')
        self.eq(s_l_country.country2iso.get('vertexLandia'), None)



import collections

import synapse.exc as s_exc

import synapse.tests.utils as s_t_utils
from synapse.tests.utils import alist

class ViewTest(s_t_utils.SynTest):

    async def test_view_nomerge(self):
        async with self.getTestCore() as core:
            view = await core.callStorm('return($lib.view.get().fork().iden)')
            opts = {'view': view}

            await core.nodes('[ ou:org=* ]', opts=opts)
            await core.nodes('$lib.view.get().set(nomerge, $lib.true)', opts=opts)

            with self.raises(s_exc.CantMergeView):
                await core.nodes('$lib.view.get().merge()', opts=opts)

            await core.nodes('$lib.view.get().set(nomerge, $lib.false)', opts=opts)
            await core.nodes('$lib.view.get().merge()', opts=opts)

            self.len(1, await core.nodes('ou:org'))

            # mop up some coverage issues
            with self.raises(s_exc.BadOptValu):
                await core.view.setViewInfo('hehe', 10)

    async def test_view_set_parent(self):

        async with self.getTestCore() as core:

            view00 = core.getView()
            view01 = core.getView((await view00.fork())['iden'])

            await view00.nodes('[ inet:fqdn=vertex.link ]')
            await view01.nodes('inet:fqdn=vertex.link [ +#foo ]')

            # one to insert on the bottom
            layr02 = await core.addLayer()
            vdef02 = {'layers': [layr02['iden']]}
            view02 = core.getView((await core.addView(vdef=vdef02))['iden'])

            # test the storm APIs for setting view parent
            opts = {'vars': {'base': view02.iden, 'fork': view00.iden}}
            await core.stormlist('$lib.view.get($fork).set(parent, $base)', opts=opts)

            # test that merging selected nodes works correctly
            self.len(0, await view02.nodes('inet:fqdn=vertex.link'))
            msgs = await view00.stormlist('inet:fqdn | merge --apply')
            self.len(1, await view02.nodes('inet:fqdn=vertex.link'))
            self.len(0, await view02.nodes('inet:fqdn=vertex.link +#foo'))

            # check that edits made to the new base layer are reflected in forks
            self.len(1, await view02.nodes('inet:fqdn=vertex.link [+#bar]'))
            self.len(1, await view00.nodes('#bar'))
            self.len(1, await view01.nodes('#bar'))

            # test that the API prevents you from setting view parent that's already set
            opts = {'vars': {'base': view02.iden, 'fork': view00.iden}}
            msgs = await core.stormlist('$lib.view.get($fork).set(parent, $base)', opts=opts)
            self.stormIsInErr('You may not set parent on a view which already has one', msgs)

            opts = {'vars': {'fork': view00.iden}}
            msgs = await core.stormlist('$lib.view.get($fork).set(parent, $lib.guid())', opts=opts)
            self.stormIsInErr('The parent view must already exist', msgs)

            opts = {'vars': {'fork': view00.iden}}
            msgs = await core.stormlist('$lib.view.get($fork).set(parent, $fork)', opts=opts)
            self.stormIsInErr('A view may not have parent set to itself', msgs)

            opts = {'vars': {'fork': view00.iden, 'base': view01.iden}}
            msgs = await core.stormlist('$lib.view.get($fork).set(parent, $base)', opts=opts)
            self.stormIsInErr('Circular dependency of view parents is not supported', msgs)

            layr03 = await core.addLayer()
            layr04 = await core.addLayer()
            vdef03 = {'layers': [layr03['iden'], layr04['iden']]}
            vdef03 = await core.addView(vdef=vdef03)

            opts = {'vars': {'fork': vdef03['iden'], 'base': view01.iden}}
            msgs = await core.stormlist('$lib.view.get($fork).set(parent, $base)', opts=opts)
            self.stormIsInErr('You may not set parent on a view which has more than one layer', msgs)

    async def test_view_fork_merge(self):

        async with self.getTestCore() as core:
            await core.nodes('[ test:int=8 +#faz ]')
            await core.nodes('[ test:int=9 test:int=10 ]')
            await core.auth.addUser('visi')
            await core.addTagProp('score', ('int', {}), {})

            self.len(1, await alist(core.eval('test:int=8 +#faz')))
            self.len(2, await alist(core.eval('test:int=9 test:int=10')))
            self.eq(3, (await core.getFormCounts()).get('test:int'))

            # Fork the main view
            vdef2 = await core.view.fork()
            view2_iden = vdef2.get('iden')
            view2 = core.getView(view2_iden)

            # The new view has the same nodes as the old view
            nodes = await alist(view2.eval('test:int=10'))
            self.len(1, nodes)

            await core.nodes('[ test:int=11 ]')

            # A node added to the parent after the fork is still seen by the child
            nodes = await alist(view2.eval('test:int=11'))
            self.len(1, nodes)
            self.eq(4, (await core.getFormCounts()).get('test:int'))

            # A node added to the child is not seen by the parent
            nodes = await alist(view2.eval('[ test:int=12 ]'))
            self.len(1, nodes)

            nodes = await core.nodes('test:int=12')
            self.len(0, nodes)
            self.eq(4, (await core.view.getFormCounts()).get('test:int'))

            # Deleting nodes from the child view should not affect the main
            await alist(view2.eval('test:int | delnode'))

            self.eq(4, (await core.view.getFormCounts()).get('test:int'))
            nodes = await alist(view2.eval('test:int=10'))
            self.len(1, nodes)

            self.eq(4, (await core.getFormCounts()).get('test:int'))
            await self.agenlen(0, view2.eval('test:int=12'))

            # Until we get tombstoning, the child view can't delete a node in the lower layer
            await self.agenlen(1, view2.eval('test:int=10'))

            # Add a node back
            await self.agenlen(1, view2.eval('[ test:int=12 ]'))

            # Add a bunch of nodes to require chunking of splices when merging
            for i in range(1000):
                await self.agenlen(1, view2.eval('[test:int=$val]', opts={'vars': {'val': i + 1000}}))

            # Add prop that will only exist in the child
            await alist(view2.eval('test:int=10 [:loc=us]'))
            self.len(1, await alist(view2.eval('test:int=10 +:loc=us')))
            self.len(0, await core.nodes('test:int=10 +:loc=us'))

            # Add tag that will only exist in child
            await alist(view2.eval('test:int=11 [+#foo.bar:score=20]'))
            self.len(1, await alist(view2.eval('test:int=11 +#foo.bar:score=20')))
            self.len(0, await core.nodes('test:int=11 +#foo.bar:score=20'))

            # Add tag prop that will only exist in child
            await alist(view2.eval('test:int=8 [+#faz:score=55]'))
            self.len(1, await alist(view2.eval('test:int=8 +#faz:score=55')))
            self.len(0, await core.nodes('test:int=8 +#faz:score=55'))

            # Add nodedata that will only exist in child
            await alist(view2.eval('test:int=9 $node.data.set(spam, ham)'))
            self.len(1, await view2.callStorm('test:int=9 return($node.data.list())'))
            self.len(0, await core.callStorm('test:int=9 return($node.data.list())'))

            # Add edges that will only exist in the child
            await alist(view2.eval('test:int=9 [ +(refs)> {test:int=10} ]'))
            await alist(view2.eval('test:int=12 [ +(refs)> {test:int=11} ]'))
            self.len(2, await alist(view2.eval('test:int -(refs)> *')))
            self.len(0, await core.nodes('test:int -(refs)> *'))

            # Forker and forkee have their layer configuration frozen
            tmplayr = await core.addLayer()
            tmpiden = tmplayr['iden']
            await self.asyncraises(s_exc.ReadOnlyLayer, core.view.addLayer(tmpiden))
            await self.asyncraises(s_exc.ReadOnlyLayer, view2.addLayer(tmpiden))
            await self.asyncraises(s_exc.ReadOnlyLayer, core.view.setLayers([tmpiden]))
            await self.asyncraises(s_exc.ReadOnlyLayer, view2.setLayers([tmpiden]))

            # You can't merge a non-forked view
            await self.asyncraises(s_exc.SynErr, view2.core.view.merge())

            # You can't merge if the parent's write layer is readonly
            view2.parent.layers[0].readonly = True
            await self.asyncraises(s_exc.ReadOnlyLayer, view2.merge())
            view2.parent.layers[0].readonly = False

            vdef3 = await view2.fork()
            view3_iden = vdef3.get('iden')
            view3 = core.getView(view3_iden)

            # You can't delete a view or merge it if it has children
            await self.asyncraises(s_exc.SynErr, view2.merge())
            await self.asyncraises(s_exc.SynErr, view2.core.delView(view2.iden))
            await self.asyncraises(s_exc.SynErr, view2.core.delView(view2.iden))
            layr = await core.addLayer()
            layriden = layr['iden']
            await self.asyncraises(s_exc.SynErr, view2.addLayer(layriden))
            await view3.core.delView(view3.iden)

            async with core.getLocalProxy(user='visi') as prox:
                with self.raises(s_exc.AuthDeny):
                    await prox.eval('test:int=12', opts={'view': view2.iden}).list()

            # The parent count is correct
            self.eq(4, (await core.view.getFormCounts()).get('test:int'))

            # Merge the child back into the parent
            await view2.merge()

            # The parent counts includes all the nodes that were merged
            self.eq(1005, (await core.view.getFormCounts()).get('test:int'))

            # A node added to the child is now present in the parent
            nodes = await core.nodes('test:int=12')
            self.len(1, nodes)

            # The child can still see the parent's pre-existing node
            nodes = await view2.nodes('test:int=10')
            self.len(1, nodes)

            # Prop that was only set in child is present in parent
            self.len(1, await core.nodes('test:int=10 +:loc=us'))
            self.len(1, await core.nodes('test:int:loc=us'))

            # Tag that was only set in child is present in parent
            self.len(1, await core.nodes('test:int=11 +#foo.bar:score=20'))
            self.len(1, await core.nodes('test:int#foo.bar'))

            # Tagprop that as only set in child is present in parent
            self.len(1, await core.nodes('test:int=8 +#faz:score=55'))
            self.len(1, await core.nodes('test:int#faz:score=55'))

            # Node data that was only set in child is present in parent
            self.len(1, await core.callStorm('test:int=9 return($node.data.list())'))
            self.len(1, await core.nodes('yield $lib.lift.byNodeData(spam)'))

            # Edge that was only set in child present in parent
            self.len(2, await core.nodes('test:int -(refs)> *'))

            # The child count includes all the nodes in the view
            self.eq(1005, (await view2.getFormCounts()).get('test:int'))

            # The child can see nodes that got merged
            nodes = await view2.nodes('test:int=12')
            self.len(1, nodes)
            nodes = await view2.nodes('test:int=1000')
            self.len(1, nodes)

            await core.delView(view2.iden)
            await core.view.addLayer(layriden)

            # But not the same layer twice
            await self.asyncraises(s_exc.DupIden, core.view.addLayer(layriden))

    async def test_view_trigger(self):
        async with self.getTestCore() as core:

            # Fork the main view
            vdef2 = await core.view.fork()
            view2_iden = vdef2.get('iden')
            view2 = core.getView(view2_iden)

            await core.view.addTrigger({
                'cond': 'node:add',
                'form': 'test:int',
                'storm': '[ test:str=mainhit ]'
            })

            nodes = await alist(core.eval('[ test:int=11 ]', opts={'view': view2.iden}))
            self.len(1, nodes)

            self.len(0, await core.view.nodes('test:str=mainhit'))

            # A trigger on the child view fires on the child view but not the main view
            await view2.addTrigger({
                'cond': 'node:add',
                'form': 'test:int',
                'storm': '[ test:str=forkhit ]',
            })

            nodes = await alist(view2.eval('[ test:int=12 ]'))

            nodes = await view2.nodes('test:str=forkhit')
            self.len(1, nodes)

            nodes = await alist(core.view.eval('test:str=forkhit'))
            self.len(0, nodes)

            # listTriggers should show only that view's triggers
            trigs = await view2.listTriggers()
            self.len(1, trigs)

            await view2.addTrigger({
                'cond': 'tag:add',
                'tag': 'foo',
                'storm': '[ +#bar ]',
            })

            await view2.addTrigger({
                'cond': 'tag:del',
                'tag': 'foo',
                'storm': '[ -#bar ]',
            })

            await view2.addTrigger({
                'cond': 'tag:add',
                'tag': 'foo',
                'storm': '| newpnewp',
            })

            await view2.addTrigger({
                'cond': 'tag:del',
                'tag': 'foo',
                'storm': '| newpnewp',
            })

            nodes = await view2.nodes('test:str=forkhit [+#foo]')
            self.len(1, nodes)
            self.nn(nodes[0].getTag('foo'))
            self.nn(nodes[0].getTag('bar'))

            nodes = await view2.nodes('test:str=forkhit [-#foo]')
            self.len(1, nodes)
            self.none(nodes[0].getTag('foo'))
            self.none(nodes[0].getTag('bar'))

            await view2.merge()

            # Trigger runs on merged nodes in main view
            self.len(1, await core.view.nodes('test:str=mainhit'))

            await view2.fini()
            await view2.delete()

    async def test_storm_editformat(self):
        async with self.getTestCore() as core:
            mesgs = await core.stormlist('[test:str=foo1 :hehe=bar]', opts={'editformat': 'nodeedits'})
            count = collections.Counter(m[0] for m in mesgs)
            self.eq(1, count['init'])
            self.eq(1, count['fini'])
            self.eq(1, count['node'])
            self.eq(2, count['node:edits'])
            self.eq(0, count['node:add'])

            mesgs = await core.stormlist('[test:str=foo2 :hehe=bar]', opts={'editformat': 'splices'})
            count = collections.Counter(m[0] for m in mesgs)
            self.eq(1, count['init'])
            self.eq(1, count['node:add'])
            self.eq(2, count['prop:set'])  # .created and .hehe
            self.eq(0, count['node:edits'])
            self.eq(1, count['node'])
            self.eq(1, count['fini'])

            mesgs = await core.stormlist('[test:str=foo3 :hehe=bar]', opts={'editformat': 'count'})
            count = collections.Counter(m[0] for m in mesgs)
            self.eq(1, count['init'])
            self.eq(1, count['node'])
            self.eq(1, count['fini'])
            self.eq(2, count['node:edits:count'])
            self.eq(0, count['node:edits'])
            self.eq(0, count['node:add'])
            cmsgs = [m[1] for m in mesgs if m[0] == 'node:edits:count']
            self.eq([{'count': 2}, {'count': 1}], cmsgs)

            mesgs = await core.stormlist('[test:str=foo3 :hehe=bar]', opts={'editformat': 'none'})
            count = collections.Counter(m[0] for m in mesgs)
            self.eq(1, count['init'])
            self.eq(0, count['node:edits:count'])
            self.eq(0, count['node:edits'])
            self.eq(0, count['node:add'])
            self.eq(1, count['node'])
            self.eq(1, count['fini'])

            with self.raises(s_exc.BadConfValu):
                await core.stormlist('[test:str=foo3 :hehe=bar]', opts={'editformat': 'jsonl'})

    async def test_lib_view_addNodeEdits(self):

        async with self.getTestCore() as core:

            view = await core.callStorm('''
                $layr = $lib.layer.add().iden
                $view = $lib.view.add(($layr,))
                return($view.iden)
            ''')

            await core.nodes('trigger.add node:add --form ou:org --query {[+#foo]}', opts={'view': view})

            nodes = await core.nodes('[ ou:org=* ]')
            self.len(0, await core.nodes('ou:org', opts={'view': view}))

            await core.stormlist('''
                $view = $lib.view.get($viewiden)
                for ($offs, $edits) in $lib.layer.get().edits(wait=$lib.false) {
                    $view.addNodeEdits($edits)
                }
            ''', opts={'vars': {'viewiden': view}})

            self.len(1, await core.nodes('ou:org +#foo', opts={'view': view}))



import regex
import asyncio

import synapse.exc as s_exc
import synapse.lib.cmdr as s_cmdr

import synapse.tests.utils as s_t_utils


class CmdBossTest(s_t_utils.SynTest):

    async def test_ps_kill(self):

        async with self.getTestCoreAndProxy() as (realcore, core):

            evnt = asyncio.Event()

            outp = self.getTestOutp()
            cmdr = await s_cmdr.getItemCmdr(core, outp=outp)

            await cmdr.runCmdLine('ps')

            self.true(outp.expect('0 tasks found.'))

            async def runLongStorm():
                async for _ in core.storm(f'[ test:str=foo test:str={"x"*100} ] | sleep 10 | [ test:str=endofquery ]'):
                    evnt.set()

            task = realcore.schedCoro(runLongStorm())

            self.true(await asyncio.wait_for(evnt.wait(), timeout=6))

            # Verify that the long query got truncated
            outp.clear()
            await cmdr.runCmdLine('ps')

            self.true(outp.expect('xxx...'))
            self.true(outp.expect('1 tasks found.'))
            self.true(outp.expect('start time: 2'))

            # Verify we see the whole query
            outp.clear()
            await cmdr.runCmdLine('ps -v')

            self.true(outp.expect('endofquery'))
            self.true(outp.expect('1 tasks found.'))
            self.true(outp.expect('start time: 2'))

            regx = regex.compile('task iden: ([a-f0-9]{32})')
            match = regx.match(str(outp))

            iden = match.groups()[0]

            outp.clear()
            await cmdr.runCmdLine('kill')
            outp.expect('Kill a running task/query within the cortex.')

            outp.clear()
            await cmdr.runCmdLine('kill %s' % (iden,))

            outp.expect('kill status: True')
            self.true(task.done())

            outp.clear()
            await cmdr.runCmdLine('ps')
            self.true(outp.expect('0 tasks found.'))

        async with self.getTestCoreAndProxy() as (realcore, core):

            bond = await realcore.auth.addUser('bond')

            async with realcore.getLocalProxy(user='bond') as tcore:

                evnt = asyncio.Event()

                async def runLongStorm():
                    async for mesg in core.storm('[ test:str=foo test:str=bar ] | sleep 10'):
                        evnt.set()

                outp = self.getTestOutp()
                cmdr = await s_cmdr.getItemCmdr(core, outp=outp)

                toutp = self.getTestOutp()
                tcmdr = await s_cmdr.getItemCmdr(tcore, outp=toutp)

                task = realcore.schedCoro(runLongStorm())
                self.true(await asyncio.wait_for(evnt.wait(), timeout=6))

                outp.clear()
                await cmdr.runCmdLine('ps')
                self.true(outp.expect('1 tasks found.'))

                regx = regex.compile('task iden: ([a-f0-9]{32})')
                match = regx.match(str(outp))
                iden = match.groups()[0]

                toutp.clear()
                await tcmdr.runCmdLine('ps')
                self.true(toutp.expect('0 tasks found.'))

                # Try killing from the unprivileged user
                await self.asyncraises(s_exc.AuthDeny, tcore.kill(iden))
                toutp.clear()
                await tcmdr.runCmdLine('kill %s' % (iden,))
                self.true(toutp.expect('no matching process found.'))

                # Try a kill with a numeric identifier - this won't match
                toutp.clear()
                await tcmdr.runCmdLine('kill 123412341234')
                self.true(toutp.expect('no matching process found', False))

                # Specify the iden arg multiple times
                toutp.clear()
                await tcmdr.runCmdLine('kill 123412341234 deadb33f')
                self.true(toutp.expect('unrecognized arguments', False))

                # Give user explicit permissions to list
                await core.addUserRule(bond.iden, (True, ('task', 'get')))

                # List now that the user has permissions
                toutp.clear()
                await tcmdr.runCmdLine('ps')
                self.true(toutp.expect('1 tasks found.'))

                # Give user explicit license to kill
                await core.addUserRule(bond.iden, (True, ('task', 'del')))

                # Kill the task as the user
                toutp.clear()
                await tcmdr.runCmdLine('kill %s' % (iden,))
                toutp.expect('kill status: True')
                self.true(task.done())



import math

'''
Synapse module with helpers for earth based geo-spacial calculations.
'''

# base earth geo distances will be in mm
r_mm = 6371008800.0
r_km = 6371.0088

# investigate perf impact of using WGS-84 ellipsoid for dist calc

def latlong(text):
    '''
    Chop a latlong string and return (float,float).
    Does not perform validation on the coordinates.

    Args:
        text (str):  A longitude,latitude string.

    Returns:
        (float,float): A longitude, latitude float tuple.
    '''
    nlat, nlon = text.split(',')
    return (float(nlat), float(nlon))

def near(point, dist, points):
    '''
    Determine if the given point is within dist of any of points.

    Args:
        point ((float,float)): A latitude, longitude float tuple.
        dist (int): A distance in mm ( base units )
        points (list): A list of latitude, longitude float tuples to compare against.
    '''
    for cmpt in points:
        if haversine(point, cmpt) <= dist:
            return True
    return False

def haversine(px, py, r=r_mm):
    '''
    Calculate the haversine distance between two points
    defined by (lat,lon) tuples.

    Args:
        px ((float,float)): lat/long position 1
        py ((float,float)): lat/long position 2
        r (float): Radius of sphere

    Returns:
        (int):  Distance in mm.
    '''
    lat1, lon1 = px
    lat2, lon2 = py

    dlat = math.radians(lat2 - lat1)
    dlon = math.radians(lon2 - lon1)
    lat1 = math.radians(lat1)
    lat2 = math.radians(lat2)

    a = math.sin(dlat / 2) ** 2 + math.cos(lat1) * math.cos(lat2) * math.sin(dlon / 2) ** 2
    c = 2 * math.asin(math.sqrt(a))

    return c * r

def bbox(lat, lon, dist):
    '''
    Calculate a min/max bounding box for the circle defined by lalo/dist.

    Args:
        lat (float): The latitude in degrees
        lon (float): The longitude in degrees
        dist (int): A distance in geo:dist base units (mm)

    Returns:
        (float,float,float,float): (latmin, latmax, lonmin, lonmax)
    '''
    latr = math.radians(lat)
    lonr = math.radians(lon)

    rad = r_mm
    prad = rad * math.cos(latr)

    latd = dist / rad
    lond = dist / prad

    latmin = math.degrees(latr - latd)
    latmax = math.degrees(latr + latd)
    lonmin = math.degrees(lonr - lond)
    lonmax = math.degrees(lonr + lond)

    return (latmin, latmax, lonmin, lonmax)

def dms2dec(degs, mins, secs):
    '''
    Convert degrees, minutes, seconds lat/long form to degrees float.

    Args:
        degs (int): Degrees
        mins (int): Minutes
        secs (int): Seconds

    Returns:
        (float): Degrees
    '''
    return degs + (mins / 60.0) + (secs / 3600.0)



import synapse.exc as s_exc
import synapse.lib.urlhelp as s_urlhelp

import synapse.tests.utils as s_t_utils

class UrlTest(s_t_utils.SynTest):
    def test_urlchop(self):

        url = 'http://vertex.link:8080/hehe.html'
        info = s_urlhelp.chopurl(url)
        self.eq({'scheme': 'http',
                 'port': 8080,
                 'host': 'vertex.link',
                 'path': '/hehe.html',
                 },
                info
                )

        url = 'tcp://pennywise:candy@vertex.link/'
        info = s_urlhelp.chopurl(url)
        self.eq({'scheme': 'tcp',
                 'user': 'pennywise',
                 'host': 'vertex.link',
                 'path': '/',
                 'passwd': 'candy',
                 },
                info
                )

        url = 'tcp://pennywise@vertex.link'
        info = s_urlhelp.chopurl(url)
        self.eq({'scheme': 'tcp',
                 'user': 'pennywise',
                 'host': 'vertex.link',
                 'path': '',
                 },
                info
                )

        url = 'tcp://1.2.3.4:8080/api/v1/wow?key=valu&foo=bar'
        info = s_urlhelp.chopurl(url)
        self.eq({'scheme': 'tcp',
                 'host': '1.2.3.4',
                 'port': 8080,
                 'path': '/api/v1/wow',
                 'query': {'key': 'valu',
                           'foo': 'bar',
                           }
                 },
                info
                )

        url = 'http://[1fff:0:a88:85a3::ac1f]:8001/index.html'
        info = s_urlhelp.chopurl(url)
        self.eq({'scheme': 'http',
                 'host': '1fff:0:a88:85a3::ac1f',
                 'port': 8001,
                 'path': '/index.html',
                 },
                info
                )

        url = 'http://::1/index.html'
        info = s_urlhelp.chopurl(url)
        self.eq({'scheme': 'http',
                 'host': '::1',
                 'path': '/index.html',
                 },
                info
                )

        self.raises(s_exc.BadUrl, s_urlhelp.chopurl,
                    'www.vertex.link')

    def test_sanitizeUrl(self):
        data = [
            ('http://foo.com/path#fragment', None),
            ('http://foo.com:1234?query=bar', None),
            ('rando:this:is:valid:URI', None),
            ('rando:this:is@valid:URI', None),
            ('foo://user:password@host.com', 'foo://user:****@host.com'),
            ('foo://user:password@host.com:999', 'foo://user:****@host.com:999'),
            ('foo://user:@host.com', None),
            ('ssl://feeds00.v.link:43/*/feed/6a1f?cere=root@.vex.link', None),
        ]

        for in_, out in data:
            self.eq(s_urlhelp.sanitizeUrl(in_), in_ if out is None else out)



import socket
import asyncio
import logging

import collections

logger = logging.getLogger(__name__)

import synapse.exc as s_exc
import synapse.common as s_common

import synapse.lib.base as s_base
import synapse.lib.const as s_const
import synapse.lib.msgpack as s_msgpack

readsize = 10 * s_const.megabyte

async def connect(host, port, ssl=None, hostname=None):
    '''
    Async connect and return a Link().
    '''
    info = {'host': host, 'port': port, 'ssl': ssl, 'hostname': hostname}
    reader, writer = await asyncio.open_connection(host, port, ssl=ssl, server_hostname=hostname)
    return await Link.anit(reader, writer, info=info)

async def listen(host, port, onlink, ssl=None):
    '''
    Listen on the given host/port and fire onlink(Link).

    Returns a server object that contains the listening sockets
    '''
    async def onconn(reader, writer):
        info = {'tls': bool(ssl)}
        link = await Link.anit(reader, writer, info=info)
        link.schedCoro(onlink(link))

    server = await asyncio.start_server(onconn, host=host, port=port, ssl=ssl)
    return server

async def unixlisten(path, onlink):
    '''
    Start an PF_UNIX server listening on the given path.
    '''
    info = {'path': path, 'unix': True}

    async def onconn(reader, writer):
        link = await Link.anit(reader, writer, info=info)
        link.schedCoro(onlink(link))
    return await asyncio.start_unix_server(onconn, path=path)

async def unixconnect(path):
    '''
    Connect to a PF_UNIX server listening on the given path.
    '''
    reader, writer = await asyncio.open_unix_connection(path=path)
    info = {'path': path, 'unix': True}
    return await Link.anit(reader, writer, info=info)

async def linkfile(mode='wb'):
    '''
    Connect a socketpair to a file-object and return (link, file).
    '''
    sock0, sock1 = socket.socketpair()

    file1 = sock1.makefile(mode)
    sock1.close()

    reader, writer = await asyncio.open_connection(sock=sock0)
    link0 = await Link.anit(reader, writer, info={'unix': True})

    return link0, file1

async def linksock():
    '''
    Connect a Link, socket pair.
    '''
    sock0, sock1 = socket.socketpair()
    reader, writer = await asyncio.open_connection(sock=sock0)
    link0 = await Link.anit(reader, writer, info={'unix': True})
    return link0, sock1

async def fromspawn(spawninfo):
    sock = spawninfo.get('sock')
    info = spawninfo.get('info', {})
    info['spawn'] = True
    reader, writer = await asyncio.open_connection(sock=sock)
    return await Link.anit(reader, writer, info=info)

class Link(s_base.Base):
    '''
    A Link() is created to wrap a socket reader/writer.
    '''
    async def __anit__(self, reader, writer, info=None):

        await s_base.Base.__anit__(self)

        self.iden = s_common.guid()

        self.reader = reader
        self.writer = writer

        self.rxqu = collections.deque()

        self.sock = self.writer.get_extra_info('socket')
        self.peercert = self.writer.get_extra_info('peercert')

        self._drain_lock = asyncio.Lock()

        if info is None:
            info = {}

        if not info.get('unix'):

            # disable nagle ( to minimize latency for small xmit )
            self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

            # enable TCP keep alives...
            self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            if hasattr(socket, 'TCP_KEEPIDLE'):
                # start sending a keep alives after 3 sec of inactivity
                self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 3)
                # send keep alives every 3 seconds once started
                self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 3)
                # close the socket after 5 failed keep alives (15 sec)
                self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 5)

        self.info = info
        self.hostname = self.info.get('hostname')

        self.unpk = s_msgpack.Unpk()

        async def fini():
            self.writer.close()
            try:
                await self.writer.wait_closed()
            except (BrokenPipeError, ConnectionResetError) as e:
                logger.debug('Link error waiting on close: %s', str(e))

        self.onfini(fini)

        if self.hostname is not None:
            if self.hostname != self.getTlsPeerCn():
                mesg = f'Expected: {self.hostname} Got: {self.getTlsPeerCn()}'
                await self.fini()
                raise s_exc.BadCertHost(mesg=mesg)

    def getTlsPeerCn(self):

        if self.peercert is None:
            return None

        for items in self.peercert.get('subject', ()):
            for name, valu in items:
                if name == 'commonName':
                    return valu

    async def getSpawnInfo(self):
        info = {}

        # selectively add info for pickle...
        if self.info.get('unix'):
            info['unix'] = True

        if self.info.get('tls'):
            info['unix'] = True
            link0, sock = await linksock()
            link0.onfini(sock.close)

            async def relay(link):
                async with link:
                    while True:
                        byts = await link.recv(1024)
                        if not byts:
                            break
                        await self.send(byts)

            self.schedCoro(relay(link0))

        else:
            sock = self.reader._transport._sock

        return {
            'info': info,
            'sock': sock,
        }

    def getAddrInfo(self):
        '''
        Get a summary of address information related to the link.
        '''
        ret = {'family': 'tcp',
               'addr': self.sock.getpeername(),
               }
        # Set family information
        if self.info.get('unix'):
            ret['family'] = 'unix'
            # Unix sockets don't use getpeername
            ret['addr'] = self.sock.getsockname()
        elif self.info.get('tls'):
            ret['family'] = 'tls'
        # Set ipver if needed
        if self.sock.family == socket.AF_INET:
            ret['ipver'] = 'ipv4'
        if self.sock.family == socket.AF_INET6:
            ret['ipver'] = 'ipv6'
        return ret

    async def send(self, byts):
        self.writer.write(byts)
        # Avoid Python bug.  See https://bugs.python.org/issue29930
        async with self._drain_lock:
            await self.writer.drain()

    async def tx(self, mesg):
        '''
        Async transmit routine which will wait for writer drain().
        '''
        if self.isfini:
            raise s_exc.IsFini()

        byts = s_msgpack.en(mesg)
        try:

            self.writer.write(byts)

            # Avoid Python bug.  See https://bugs.python.org/issue29930
            async with self._drain_lock:
                await self.writer.drain()

        except (asyncio.CancelledError, Exception) as e:

            await self.fini()

            einfo = s_common.retnexc(e)
            logger.debug('link.tx connection trouble %s', einfo)

            raise

    async def recv(self, size):
        return await self.reader.read(size)

    async def recvsize(self, size):
        byts = b''
        while size:

            recv = await self.reader.read(size)
            if not recv:
                await self.fini()
                return None

            size -= len(recv)
            byts += recv

        return byts

    async def rx(self):

        while not self.rxqu:

            if self.isfini:
                return None

            try:

                byts = await self.reader.read(readsize)
                if not byts:
                    await self.fini()
                    return None

                for _, mesg in self.feed(byts):
                    self.rxqu.append(mesg)

            except asyncio.CancelledError:
                await self.fini()
                raise

            except Exception as e:
                if isinstance(e, (BrokenPipeError, ConnectionResetError)):
                    logger.warning('rx error: %s', e)
                else:
                    logger.exception('rx error')
                await self.fini()
                return None

        return self.rxqu.popleft()

    def get(self, name, defval=None):
        '''
        Get a property from the Link info.
        '''
        return self.info.get(name, defval)

    def set(self, name, valu):
        '''
        Set a property in the Link info.
        '''
        self.info[name] = valu

    def feed(self, byts):
        '''
        Used by Plex() to unpack bytes.
        '''
        return self.unpk.feed(byts)



import synapse.exc as s_exc
import synapse.lib.chop as s_chop

import synapse.tests.utils as s_t_utils

class ChopTest(s_t_utils.SynTest):
    def test_chop_digits(self):
        self.eq(s_chop.digits('a1b2c3'), '123')

    def test_chop_tags(self):
        tags = s_chop.tags('foo.bar.baz')
        self.eq(tags, ('foo', 'foo.bar', 'foo.bar.baz'))

    def test_chop_tag(self):
        self.eq('foo.bar.ba z', s_chop.tag('#foo  .bar.  BA Z'))

    def test_chop_hexstr(self):
        testvectors = [
            ('0C', '0c'),
            ('10', '10'),
            ('0xffff', 'ffff'),
            ('0x0001', '0001'),
            ('C', s_exc.BadTypeValu),
            ('0xfff', s_exc.BadTypeValu),
            ('10001', s_exc.BadTypeValu),
            ('', s_exc.BadTypeValu),
            ('0x', s_exc.BadTypeValu),
            ('newp', s_exc.BadTypeValu)
        ]

        for v, e in testvectors:
            if isinstance(e, str):
                self.eq(s_chop.hexstr(v), e)
            else:
                self.raises(e, s_chop.hexstr, v)

    def test_chop_onespace(self):
        tvs = [
            ('asdfasdf  asdfasdf ', 'asdfasdf asdfasdf'),
            ('asdfasdf ', 'asdfasdf'),
            ('asdf', 'asdf'),
            ('  asdfasdf  ', 'asdfasdf'),
            (' asdf  asdf    asdf \t \t asdf asdf   ', 'asdf asdf asdf asdf asdf'),
            (' ', ''),
            ('foo   bar baz', 'foo bar baz'),
        ]
        for iv, ev in tvs:
            rv = s_chop.onespace(iv)
            self.eq(rv, ev)

    def test_chop_printables(self):
        tvs = [
            ('hehe haha', 'hehe haha'),
            ('hehe\u200bhaha\u200b ', 'hehehaha ')
        ]
        for iv, ev in tvs:
            rv = s_chop.printables(iv)
            self.eq(rv, ev)

    def test_chop_stormstring(self):
        tvs = [
            ('', ''),  # no change
            ('beep', 'beep'),  # no change
            ('''be'ep''', '''be'ep'''),  # no change
            ('beep"', 'beep\\"'),  # Simple case
            ('be"ep', 'be\\"ep'),  # Simple case
            ('"', '\\"'),
            ('""', '\\"\\"'),
            ('\\"', '\\\\\\"'),
            ('"\\""', '\\"\\\\\\"\\"'),
            ('\\', '\\\\'),
            ('be\\"ep', 'be\\\\\\"ep'),
        ]
        for tv, ev in tvs:
            gv = s_chop.stormstring(tv)
            self.eq(gv, ev)

    def test_chop_validatetagmatch(self):
        self.raises(s_exc.BadTag, s_chop.validateTagMatch, ' foo')
        self.raises(s_exc.BadTag, s_chop.validateTagMatch, 'foo&bar')
        self.raises(s_exc.BadTag, s_chop.validateTagMatch, 'foo..bar')
        self.none(s_chop.validateTagMatch('foo.*.bar'))
        self.none(s_chop.validateTagMatch('**foo.*.bar'))
